;====================================================================
; NVIDIA NForce2 custom option ROM for DFI Lanparty Ultra B
; vB.2A
; author: digitalbath
; Jan. 2021
; Credits to tictac, pinczakko, polygon, TzK and Infrared
; Version History
; vB.2A: new Version 
;====================================================================
use16                                           ; 16bit mode
ROM_SIZE_IN_BLOCK = 4                           ; 1 means ROM size is 1 block (512 bytes)
ROM_SIZE_IN_BYTE = ROM_SIZE_IN_BLOCK * 512      ; number of 512 byte blocks

define PCI_ADDR_PORT    0xCF8
define PCI_DATA_PORT    0xCFC
define CMOS_ADDR_PORT   0x70
define CMOS_DATA_PORT   0x71

macro configurePort pci
{
    mov eax, pci                                ; move address for 32bit register offset
    mov dx, PCI_ADDR_PORT                       ; pci register address port
    out dx, eax                                 ; set port we want to read
}

macro readPciReg pci
{
    configurePort pci                           ; pci register address port
    mov dx, PCI_DATA_PORT                       ; pci register data port
    in eax, dx                                  ; read register values
}

macro writePciReg pci, data, mask
{
    readPciReg pci                              ; pci register data port
    and eax, mask                               ; mask for data to remove pre-set value
    or eax, data                                ; add custom value to old register data
    out dx, eax                                 ; write new custom register value
}

macro readCmosReg reg, mask
{
    mov al, reg                                 ; set CMOS index
    out CMOS_ADDR_PORT, al                      ; send register offset
    in al, CMOS_DATA_PORT                       ; read CMOS data
    and al, mask                                ; mask the data and zero unneeded bits
}

ROMStart:
    db 0x055, 0x0AA                             ; ROM Header 55,AA -> Bootable rom
    db (ROMEnd - ROMStart)/512                  ; ROM Size in 512 bytes
    jmp MAIN                                    ; jump to main routine
    db 0                                        ; checksum, to be filled in later
    times (256)-($-$$) db 0                     ; locate Main routine at 100h


MAIN:
    ; save all registers for later restore
    pushfd
    pusha
;
;====================================================================
;
; CMOS register and mask used in _EN_CODE.bin and system module:
;
;
; Cmos register 78h 0x19h  T(ref)   Bus: 0 Device: 0 Function: 1 Mem Controller
; Register 60h+61h 16-bit
;
;   00: Auto (keep value, skip custom code)
;   01: 0402 -> 0204h =  516 cycles
;   02: 8802 -> 0288h =  648 cycles
;   03: 0803 -> 0308h =  776 cycles
;   04: 0D03 -> 030Dh =  780 cycles
;   05: 8C03 -> 038Ch =  908 cycles
;   06: 0804 -> 0408h = 1032 cycles
;   07: 9004 -> 0490h = 1168 cycles
;   08: 1005 -> 0510h = 1296 cycles
;   09: 0006 -> 0600h = 1536 cycles
;   0A: 1006 -> 0610h = 1552 cycles
;   0B: 1806 -> 0618h = 1560 cycles
;   0C: 1807 -> 0718h = 1816 cycles
;   0D: 0008 -> 0800h = 2048 cycles
;   0E: 1008 -> 0810h = 2064 cycles
;   0F: 2009 -> 0920h = 2336 cycles
;   10: 000A -> 0A00h = 2560 cycles
;   11: 200A -> 0A20h = 2592 cycles
;   12: 000C -> 0C00h = 3072 cycles
;   13: 300C -> 0C30h = 3120 cycles
;   14: 300E -> 0E30h = 3632 cycles
;   15: 640E -> 0E64h = 3684 cycles
;   16: 2010 -> 1020h = 4128 cycles
;   17: 6410 -> 1064h = 4196 cycles
;   18: 4012 -> 1240h = 4672 cycles
;   19: 6412 -> 1264h = 4708 cycles
;
;
; Cmos register 73h 0x1Fh  romsips tables  Bus: 0 Device: 0 Function: 3 Mem Controller
;   8-bit Register 65h, 66h, 68h, 69h, 6Ah, 6Bh, 6Ch, 6Dh, 6Fh
;   0: Auto (keep value, skip custom code)
;   1: A DFI 12/10:  65h:06 66h:F7 68h:75 69h:00 6Ah:55 6Bh:00 6Ch:66 6Dh:55 6Fh:33
;   2: B DFI 12/18:  65h:06 66h:F7 68h:55 69h:00 6Ah:66 6Bh:00 6Ch:55 6Dh:55 6Fh:33
;   3: C DFI 12/31:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   4: D DFI 01/21:  65h:06 66h:F7 68h:66 69h:00 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:F0
;   5: E DFI 01/31:  65h:00 66h:F7 68h:33 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   6: F DFI 05/05:  65h:00 66h:F7 68h:55 69h:06 6Ah:55 6Bh:00 6Ch:66 6Dh:00 6Fh:00
;   7: G custom#1 :  65h:66 66h:F7 68h:22 69h:66 6Ah:55 6Bh:00 6Ch:66 6Dh:66 6Fh:33
;   8: H custom#2 :  65h:66 66h:77 68h:22 69h:07 6Ah:97 6Bh:00 6Ch:88 6Dh:88 6Fh:77
;   9: I custom#3 :  65h:44 66h:22 68h:22 69h:22 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:22
;   A: J custom#4 :  65h:44 66h:44 68h:44 69h:44 6Ah:44 6Bh:00 6Ch:44 6Dh:44 6Fh:44
;   B: K custom#5 :  65h:66 66h:F7 68h:33 69h:66 6Ah:66 6Bh:00 6Ch:66 6Dh:66 6Fh:33   7Ch:44 48h:69 50h:B2
;   C: L custom#6 :  65h:66 66h:F7 68h:33 69h:66 6Ah:67 6Bh:00 6Ch:66 6Dh:66 6Fh:77   7Ch:44 48h:69 50h:B2
;   D:
;   E:
;   F:
;
;
;====================================================================

    readCmosReg 0x78, 0x1F                      ; read tREF from CMOS

    cmp al, 0
    je tref_end                                 ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .tref1                                   ;jump if equal 0000.0001
    cmp al, 0x02
    je .tref2                                   ;jump if equal 0000.0010
    cmp al, 0x03
    je .tref3                                   ;jump if equal 0000.0011
    cmp al, 0x04
    je .tref4                                   ;jump if equal 0000.0100
    cmp al, 0x05
    je .tref5                                   ;jump if equal 0000.0101
    cmp al, 0x06
    je .tref6                                   ;jump if equal 0000.0110
    cmp al, 0x07
    je .tref7                                   ;jump if equal 0000.0111
    cmp al, 0x08
    je .tref8                                   ;jump if equal 0000.1000
    cmp al, 0x09
    je .tref9                                   ;jump if equal 0000.1001
    cmp al, 0x0A
    je .tref10                                  ;jump if equal 0000.1010
    cmp al, 0x0B
    je .tref11                                  ;jump if equal 0000.1011
    cmp al, 0x0C
    je .tref12                                  ;jump if equal 0000.1100
    cmp al, 0x0D
    je .tref13                                  ;jump if equal 0000.1101
    cmp al, 0x0E
    je .tref14                                  ;jump if equal 0000.1110
    cmp al, 0x0F
    je .tref15                                  ;jump if equal 0000.1111
    cmp al, 0x10
    je .tref16                                  ;jump if equal 0001.0000
    cmp al, 0x11
    je .tref17                                  ;jump if equal 0001.0001
    cmp al, 0x12
    je .tref18                                  ;jump if equal 0001.0010
    cmp al, 0x13
    je .tref19                                  ;jump if equal 0001.0011
    cmp al, 0x14
    je .tref20                                  ;jump if equal 0001.0100
    cmp al, 0x15
    je .tref21                                  ;jump if equal 0001.0101
    cmp al, 0x16
    je .tref22                                  ;jump if equal 0001.0110
    cmp al, 0x17
    je .tref23                                  ;jump if equal 0001.0111
    cmp al, 0x18
    je .tref24                                  ;jump if equal 0001.1000
    cmp al, 0x19
    je .tref25                                  ;jump if equal 0001.1001
    

  .tref1:
    mov ebx, 0x0204                             ; 0516 Cycles
    je .end2
  .tref2:
    mov ebx, 0x0288                             ; 0648 Cycles
    je .end2
  .tref3:
    mov ebx, 0x0308                             ; 0776 Cycles
    je .end2
  .tref4:
    mov ebx, 0x030C                             ; 0780 Cycles
    je .end2
  .tref5:
    mov ebx, 0x038C                             ; 0908 Cycles
    je .end2
  .tref6:
    mov ebx, 0x0408                             ; 1032 Cycles
    je .end2
  .tref7:
    mov ebx, 0x0490                             ; 1168 Cycles
    je .end2
  .tref8:
    mov ebx, 0x0510                             ; 1296 Cycles
    je .end2
  .tref9:
    mov ebx, 0x0600                             ; 1536 Cycles
    je .end2
  .tref10:
    mov ebx, 0x0610                             ; 1552 Cycles
    je .end2
  .tref11:
    mov ebx, 0x0618                             ; 1560 Cycles
    je .end2
  .tref12:
    mov ebx, 0x0718                             ; 1816 Cycles
    je .end2
  .tref13:
    mov ebx, 0x0800                             ; 2048 Cycles
    je .end2
  .tref14:
    mov ebx, 0x0810                             ; 2064 Cycles
    je .end2
  .tref15:
    mov ebx, 0x0920                             ; 2336 Cycles
    je .end2
  .tref16:
    mov ebx, 0x0A00                             ; 2560 Cycles
    je .end2
  .tref17:
    mov ebx, 0x0A20                             ; 2592 Cycles
    je .end2
  .tref18:
    mov ebx, 0x0C00                             ; 3072 Cycles
    je .end2
  .tref19:
    mov ebx, 0x0C30                             ; 3120 Cycles
    je .end2
  .tref20:
    mov ebx, 0x0E30                             ; 3632 Cycles
    je .end2
  .tref21:
    mov ebx, 0x0E64                             ; 3684 Cycles
    je .end2
  .tref22:
    mov ebx, 0x1020                             ; 4128 Cycles
    je .end2
  .tref23:
    mov ebx, 0x1064                             ; 4196 Cycles
    je .end2
  .tref24:
    mov ebx, 0x1240                             ; 4672 Cycles
    je .end2
  .tref25:
    mov ebx, 0x1264                             ; 4708 Cycles
    je .end2

  .end2:
    writePciReg 0x80000160, ebx, 0xFFFF0000
    je tref_end

    tref_end:                                   ; jump target

;====================================================================

    readCmosReg 0x73, 0x1F                      ; read romsips from CMOS

    cmp al, 0
    je codeend                                  ; jump to end if auto selected (equal 0000.0000)

    xor ebx, ebx

    cmp al, 0x01
    je .sips1                                   ;jump if equal 0000.0001
    cmp al, 0x02
    je .sips2                                   ;jump if equal 0000.0010
    cmp al, 0x03
    je .sips3                                   ;jump if equal 0000.0011
    cmp al, 0x04
    je .sips4                                   ;jump if equal 0000.0100
    cmp al, 0x05
    je .sips5                                   ;jump if equal 0000.0101
    cmp al, 0x06
    je .sips6                                   ;jump if equal 0000.0110
    cmp al, 0x07
    je .sips7                                   ;jump if equal 0000.0111
    cmp al, 0x08
    je .sips8                                   ;jump if equal 0000.1000
    cmp al, 0x09
    je .sips9                                   ;jump if equal 0000.1001
    cmp al, 0x0A
    je .sips10                                  ;jump if equal 0000.1010
    cmp al, 0x0B
    je .sips11                                  ;jump if equal 0000.1011
    cmp al, 0x0C
    je .sips12                                  ;jump if equal 0000.1100
    cmp al, 0x0D
    je .sips13                                  ;jump if equal 0000.1101
    cmp al, 0x0E
    je .sips14                                  ;jump if equal 0000.1110
    cmp al, 0x0F
    je .sips15                                  ;jump if equal 0000.1111

  .sips1:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00550075, 0x00000000     ;6Bh:00 6Ah:55 69h:00 68h:75   
    writePciReg 0x8000036C, 0x33005566, 0x00FF0000     ;6Fh:33        6Dh:55 6Ch:66  
    jmp codeend
  .sips2:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660055, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:55   
    writePciReg 0x8000036C, 0x33005555, 0x00FF0000     ;6Fh:33        6Dh:55 6Ch:55  
    jmp codeend
  .sips3:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660066, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:66   
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66  
    jmp codeend
  .sips4:
    writePciReg 0x80000364, 0x00F70600, 0xFF0000FF     ;       66h:F7 65h:06
    writePciReg 0x80000368, 0x00660066, 0x00000000     ;6Bh:00 6Ah:66 69h:00 68h:66   
    writePciReg 0x8000036C, 0xF0006666, 0x00FF0000     ;6Fh:F0        6Dh:66 6Ch:66  
    jmp codeend
  .sips5:
    writePciReg 0x80000364, 0x00F70000, 0xFF0000FF     ;       66h:F7 65h:00
    writePciReg 0x80000368, 0x00550633, 0x00000000     ;6Bh:00 6Ah:55 69h:06 68h:33   
    writePciReg 0x8000036C, 0x00000066, 0x00FF0000     ;6Fh:00        6Dh:00 6Ch:66  
    jmp codeend
  .sips6:
    writePciReg 0x80000364, 0x00F70000, 0xFF0000FF     ;       66h:F7 65h:00
    writePciReg 0x80000368, 0x00550655, 0x00000000     ;6Bh:00 6Ah:55 69h:06 68h:55   
    writePciReg 0x8000036C, 0x00000066, 0x00FF0000     ;6Fh:00        6Dh:00 6Ch:66  
    jmp codeend
  .sips7:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66
    writePciReg 0x80000368, 0x00556622, 0x00000000     ;6Bh:00 6Ah:55 69h:66 68h:22   
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66  
    jmp codeend
  .sips8:
    writePciReg 0x80000364, 0x00776600, 0xFF0000FF     ;       66h:77 65h:66
    writePciReg 0x80000368, 0x00970722, 0x00000000     ;6Bh:00 6Ah:97 69h:07 68h:22   
    writePciReg 0x8000036C, 0x77008888, 0x00FF0000     ;6Fh:77        6Dh:88 6Ch:88  
    jmp codeend
  .sips9:
    writePciReg 0x80000364, 0x00224400, 0xFF0000FF     ;       66h:22 65h:44
    writePciReg 0x80000368, 0x00442222, 0x00000000     ;6Bh:00 6Ah:44 69h:22 68h:22   
    writePciReg 0x8000036C, 0x22004444, 0x00FF0000     ;6Fh:22        6Dh:44 6Ch:44  
    jmp codeend
  .sips10:
    writePciReg 0x80000364, 0x00444400, 0xFF0000FF     ;       66h:44 65h:44
    writePciReg 0x80000368, 0x00444444, 0x00000000     ;6Bh:00 6Ah:44 69h:44 68h:44   
    writePciReg 0x8000036C, 0x44004444, 0x00FF0000     ;6Fh:44        6Dh:44 6Ch:44  
    jmp codeend
  .sips11:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66  
    writePciReg 0x80000368, 0x00666633, 0x00000000     ;6Bh:00 6Ah:66 69h:66 68h:33     
    writePciReg 0x8000036C, 0x33006666, 0x00FF0000     ;6Fh:33        6Dh:66 6Ch:66 
    writePciReg 0x8000037C, 0x00000044, 0xFFFFFF00     ;                     7Ch:44
    writePciReg 0x80000348, 0x00000069, 0xFFFFFF00     ;                     48h:69
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ;                     50h:B2
    jmp codeend
  .sips12:
    writePciReg 0x80000364, 0x00F76600, 0xFF0000FF     ;       66h:F7 65h:66  
    writePciReg 0x80000368, 0x00676633, 0x00000000     ;6Bh:00 6Ah:67 69h:66 68h:33     
    writePciReg 0x8000036C, 0x77006666, 0x00FF0000     ;6Fh:77        6Dh:66 6Ch:66
    writePciReg 0x8000037C, 0x00000044, 0xFFFFFF00     ;                     7Ch:44
    writePciReg 0x80000348, 0x00000069, 0xFFFFFF00     ;                     48h:69
    writePciReg 0x80000350, 0x000000B2, 0xFFFFFF00     ;                     50h:B2 
    jmp codeend
  .sips13:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h:     
;                                                      ;6Fh:          6Dh:   6Ch:    
    jmp codeend
  .sips14:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h:     
;                                                      ;6Fh:          6Dh:   6Ch:    
    jmp codeend
  .sips15:
;                                                      ;       66h:   65h:  
;                                                      ;6Bh:   6Ah:   69h:   68h:     
;                                                      ;6Fh:          6Dh:   6Ch:  
    jmp codeend

;====================================================================

codeend:
    popa
    popfd
    retf                                        ; return far to system bios routine

                                        
  times (ROM_SIZE_IN_BYTE-$) db 0               ; use 00h as the padding bytes until we reach the ROM size
                                                ; The last byte (512th) will be the patch_byte for the checksum
                                                ; patch_byte is calculated and automagically inserted below
  PREV_CHKSUM = 0
  repeat $
  load CHKSUM byte from %-1
  CHKSUM = (PREV_CHKSUM + CHKSUM) mod 0x100
  PREV_CHKSUM = CHKSUM
  end repeat
  
  store byte (0x100 - CHKSUM) at ($-1)          ; store the patch_byte
ROMEnd:
